/* vi:set ts=8 sts=4 sw=4:
 *
 * Copyright (C) 1998,1999,2000 by Marcin Dalecki <dalecki@cs.net.pl>
 *
 * This file is part of the oki4linux.
 *
 * MARCIN DALECKI ASSUMES NO RESPONSIBILITY FOR THE USE OR INABILITY TO USE ANY
 * OF THIS SOFTWARE. THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 * KIND, AND MARCIN DALECKI EXPRESSLY DISCLAIMS ALL IMPLIED WARRANTIES,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*
 * This is the driver for the OKIPAGE 4w and OKIPAGE 4w plus printers.  Despite
 * that this is an ordinary user space program I'm calling it a driver, since
 * it's only purpose is to control the printers operation.  It can't be run in
 * an ordinary fashion, since it's employing RT system's facilities.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>

#include <ctype.h>

#define VERSION "2.0"
#define PROGRAM "oki4drv: "

/* There is a physical limit on the scan line size used by the printer.
 */
#define RASTER_LINE_SIZE    5120

/*
 * Paper size information for interfacing with ghostscripts bit driver.
 *
 * Those are the formats both supported by gs and the printer itself.  Tought
 * in fact it wouldn't matter to use some arbitrary format on the printer side.
 * About this we will care later in version of this program to come.
 */

struct size_data_t {
    char *name;
    int width;			/* in bytes */
    int height;			/* in pixels */
    unsigned char buf[2];	/* information for the printer */
} size_data[] = {

    {
	"a4", 620, 7017, {
	    0, 26
	}
    },
    {
	"a5", 439, 4958, {
	    0, 25
	}
    },
    {
	"a6", 310, 3508, {
	    0, 24
	}
    },
    {
	"b5", 522, 5908, {
	    0, 100
	}
    },
    {
	"letter", 638, 6600, {
	    0, 2
	}
    },
    {
	"legal", 638, 8400, {
	    0, 3
	}
    },
    {
	NULL, 0, 0
    },
};

/*
 * Global data set up by command line options.
 */
static int verbose = 0;
static int size_index = 4;	/* Letter paper is default */
static char *output_name = NULL;	/* default output device file */

static enum {
    manual, automatic
} paper_source = automatic;

static enum {
    text, graphics
} render_mode = graphics;


/* Darkness information send to the printer */
static unsigned char darkness[5] = {2, 1, 0, 3, 4};
static int darkness_index = 0;	/* in the range from -2 to 2 */

/* Image size command */
static unsigned char imagesize[6] = "\033!A\000\000\000";

/* Paper weight information send to the printer */
static unsigned char weight[5] = {18, 17, 16, 19, 20};
static int weight_index = 1;	/* in the range from -2 to 2 */

/*
 * Compress a raster line in compression mode 2
 */
static int
mode2_compress(unsigned char *buffer, unsigned char *buffout, int emitbytes)
{
    unsigned char *pos, *ppos, *litpos, *upper;
    int lit, i, pack;
    char last, c;

    /* trap empty lines (otherwise problems with Packbits will occure) */
    if (emitbytes == 0)
	return (0);

    /* Use Packbits compression (mode 2) on raster line */
    pos = buffer;
    litpos = buffer;
    pack = 1;
    ppos = buffout;
    last = *pos++;
    upper = buffer + emitbytes;
    while (pos < upper) {
	c = *pos++;
	if (c == last) {
	    pack++;
	    /* At least three equal bytes ==> write out preceding literal sequence */
	    if ((pack == 3) && ((lit = (int) (pos - litpos - 3)) != 0)) {
		while (lit >= 128) {
		    *ppos++ = 127;
		    for (i = 1; i <= 128; i++)
			*ppos++ = *litpos++;
		    lit -= 128;
		}
		if (lit != 0) {
		    *ppos++ = lit - 1;
		    while ((lit--) > 0)
			*ppos++ = *litpos++;
		    /* litpos is now pos-3 (i.e. first of the sequence) */
		}
	    }
	} else {
	    while (pack >= 128) {
		*ppos++ = 129;
		*ppos++ = last;
		litpos += 128;
		pack -= 128;
	    }
	    /* At least 3 equal bytes or 2 that don't break a literal sequence
	       ==> write out packed sequence */
	    if ((pack >= 3) || ((pack == 2) && (litpos == pos - 3))) {
		*ppos++ = 257 - pack;
		*ppos++ = last;
		litpos += pack;
	    }
	    pack = 1;
	    last = c;
	}
#if 0
	if (ppos > (buffout + BUFFSIZE / 2 - 129)) {
	    fprintf(ERR_STREAM,
		    "Can't compact raster character; rerun without compression!\n");
	    return (0);
	}
#endif
    }

    while (pack >= 128) {
	*ppos++ = 129;
	*ppos++ = last;
	litpos += 128;
	pack -= 128;
    }
    if ((pack >= 3) || ((pack == 2) && (litpos == pos - 3))) {
	*ppos++ = 257 - pack;
	*ppos++ = last;
	litpos += pack;
    } else if ((lit = (int) (pos - litpos)) != 0) {
	while (lit >= 128) {
	    *ppos++ = 127;
	    for (i = 1; i <= 128; i++)
		*ppos++ = *litpos++;
	    lit -= 128;
	}
	if (lit != 0) {
	    *ppos++ = lit - 1;
	    while ((lit--) > 0)
		*ppos++ = *litpos++;
	}
    }

    return ((int) (ppos - buffout));
}

/*
 * Actually write the output data to the printer.
 *
 * Here we try to control the printers ability to eat the
 * data.
 */
static void write_data(int tmput, int output)
{
    struct stat stat_buf;
    off_t size;
    unsigned char *data;
    struct sched_param params;

    errno = 0;

    /* Increse our priority, since we wan't to make sure that no other
     * application comes in our way.
     */

    setpriority(PRIO_PGRP, 0, -20);

    if (errno) {
	perror(PROGRAM "Gaining CPU priority");
	exit(EXIT_FAILURE);
    }

    errno = 0;
    fstat(tmput, &stat_buf);
    if (errno) {
	perror(PROGRAM "Stating tmp file");
	exit(EXIT_FAILURE);
    }

    size = stat_buf.st_size;

    if (!size) return;

    errno = 0;
    data = mmap(NULL, (int) size, PROT_READ,MAP_SHARED, tmput, 0);
    if (data == MAP_FAILED) {
	perror(PROGRAM "Getting tmp data filed");
	exit(0);
    }
    mlock(data, size);
    params.sched_priority = 1;
    sched_setscheduler(0, SCHED_FIFO, &params);

    if (verbose) {
      printf(" ...writing %d bytes\n", size);
    }

    /* Start drawing in paper.
     */
    write(output, "\033$A", 3);

    /* Information about the expected data amount. */
    imagesize[3] = (unsigned char) (size & 0xFF);
    imagesize[4] = (unsigned char) ((size >> 8) & 0xFF);
    imagesize[5] = (unsigned char) ((size >> 16) & 0xFF);
    write(output, imagesize, 6);

    /* Raster image Y offset setting - this is marking the margin start. */
    write(output, "\033*B\010\000", 5);

    /* Write the actual data */
    write(output, data, size);

    write(output, "\033$B", 3);	/* eject */

    //also resets, which we don't want:
    //write(output, "\033\x7F\000", 3); /* flush printer buffer */

    params.sched_priority = 0;
    sched_setscheduler(0, SCHED_OTHER, &params);
    munlock(data, size);
    munmap(data, size);
}

int  process_data(FILE *in, int tmput, int scan_width, int scan_height)
{
    int i;
    unsigned char buf[RASTER_LINE_SIZE];

    for (i = 0; i < scan_height; ++i) {
	int j;

	/* as usuall, we better play it safe... */
	memset(buf, 0, RASTER_LINE_SIZE);
	j = fread(buf, scan_width, 1, in);

	if (ferror(in)) {
	    fprintf(stderr, PROGRAM "Reading data from pipe failed.\n");
	    exit(EXIT_FAILURE);
	}

	if (j == 0)
	    break;
	else if (j != 1)
	    return j;

	/*
	 * Check whatever this is an empty line.
	 */
	for (j = 0; j < scan_width; ++j)
	    if (buf[j])
		break;

	if (j == scan_width) {
	    /* Startegy for writing empty lines. */

	    /* Raster image Y offset setting */
	    write(tmput, "\033*B\001\000", 5);
	} else {
	    unsigned char comp[2 * RASTER_LINE_SIZE];
	    /* Strategy for writing lines with content. */

	    int comp_len;
	    int trail;

	    /*
	     * Skip trailing zeros
	     */
	    trail = scan_width;
	    while ((trail > 1) && (buf[trail - 1] == 0))
		--trail;

	    comp_len = mode2_compress(buf, comp + 6, trail);
	    /* Raster data transfer */
	    comp[0] = '\033';
	    comp[1] = '*';
	    comp[2] = 'A';
	    comp[3] = '\002';
	    comp[4] = (unsigned char) ((comp_len + 2)& 0xFF);
	    comp[5] = (unsigned char) (((comp_len + 2)>> 8) & 0xFF);
	    comp_len += 6;

	    /*
	     * HACK ALLERT: I really don't understand why this seems to be
	     * necessary here. This is mimmicking two zeros at the end of
	     * the lines. Otherwise the rendering engine in the printer
	     * appears to assume, that we are duplicating the previous
	     * line!
	     */

	    comp[comp_len] = 255;
	    comp[comp_len + 1] = 0;
	    comp_len += 2;
	    /* This is slow and could be wasly improved in
	     * terms of performance.
	     */
	    write(tmput, comp, comp_len);
	}
    }

    return 1;
}

/*
 * Read data from the ghostscript rasterizer subprocess.
 */
static char *
read_data(FILE * in, char *temp_name, int output)
{
    unsigned char buf[8];
    int scan_width = size_data[size_index].width;
    int scan_height = size_data[size_index].height;
    int tmput = 0;

    /* Write the universal exit command.
     */
    write(output, "\033&B\033\%-98765X", 12);

    /* Write the paper weight information.
     */
    write(output, "\034\024\003Ai", 5);
    write(output, &weight[weight_index + 2], 1);

    /* Write darkness information.
     */
    write(output, "\034\024\007Ba", 5);
    write(output, &darkness[darkness_index + 2], 1);

    /* Write render type information.
     */
    write(output, "b\002c", 3);
    if (render_mode == text)
	buf[0] = 000;
    else
	buf[0] = 001;
    write(output, buf, 1);

    /* Write some unresolved information.
     */
    write(output, "\034\177\000", 3);

    /* Write paper source information
     */
    write(output, "\033&A\001\000", 5);
    if (paper_source == automatic)
	buf[0] = 001;
    else
	buf[0] = 002;
    write(output, buf, 1);

    /* Write paper size information.
     */
    write(output, "\033&A\002\000", 5);
    write(output, size_data[size_index].buf, 2);

    /*
     * Write information about the number of copies.
     * Actually we always write only one copy.
     */
    write(output, "\033&A\003\000\000\000\001", 8);

    /*
     * Write resolution data. We stick currently to 600x600dpi, since
     * there appears to be no need to artifically make the printer
     * perform worser then it can. (Aka: raster data format information)
     */
    write(output, "\033&A\007\000\000\000\000\000", 9);
    write(output, "\007\007", 2);
    write(output, "\000", 1);

    do {
	int j; /* raster data */

	errno = 0;
	tmput = creat(temp_name, 0600);

	if (errno) {
	    perror(PROGRAM "Creating temporary file");
	    exit(EXIT_FAILURE);
	}

	j = process_data(in, tmput, scan_width, scan_height);

	if (j == 0) {
	    /* OK we have reached the end of our printing process...
	     */
	    break;
	} else if (j != 1) {
	    fprintf(stderr, PROGRAM "Not enought data from gs.\n");
	    exit(EXIT_FAILURE);
	}

	close(tmput);
	tmput = open(temp_name, O_RDONLY);
	/* Now finally throw it out onto the printer in one chunk. */
	if (verbose) { printf("writing data from %s to device\n", temp_name); }
	write_data(tmput, output);

	close(tmput);
	unlink(temp_name);
    } while (!feof(in));

    /* Write the universal exit command, after the job. */
    write(output, "\033\%-98765X", 9);

    return temp_name;
}


/*
 * Compose the gs command line used to invoke our child.
 */
static char *compose_gs_command(char *name, int size_index)
{
    char *chunk1 = "/usr/bin/gs -q -dSAFER -dNOPAUSE -dBATCH -sDEVICE=bit -r600x600 -sPAPERSIZE=";
    char *chunk2 = " -sOutputFile=- -q ";
    char *result;

    size_t len = strlen(chunk1) +
    strlen(size_data[size_index].name) +
    strlen(chunk2) +
    strlen(name);

    result = (char *) malloc(len + 1);
    strcpy(result, chunk1);
    strcat(result, size_data[size_index].name);
    strcat(result, chunk2);
    strcat(result, name);

    if (verbose) {
	puts("GhostScript command used:");
	printf("[%s]\n", result);
    }
    return result;
}


/*
 * Main working horse procedure.
 */
static void process_file(char *name)
{
    FILE *in;
    int output;
    char *command;
    char *temp_name;

    if (verbose)
	puts("Starting GhostScript...");

    output = open(output_name, O_CREAT | O_RDWR | O_TRUNC, 0600);
    if (output == -1) {
	fprintf(stderr, PROGRAM "Couldn't open output file %s\n", output_name);
	exit(EXIT_FAILURE);
    }
    command = compose_gs_command(name, size_index);
    in = popen(command, "r");
    if (!in) {
	fprintf(stderr, PROGRAM "Could't run gs. Please make sure it's in $PATH\n");
	exit(EXIT_FAILURE);
    }
    if (verbose)
	printf("Processing file: %s\n", name);

    temp_name = tmpnam(NULL);
    read_data(in, temp_name, output);

    free(command);
    pclose(in);
    close(output);
}

int main(int argc, char *argv[])
{
    int index;
    int c;

    opterr = 0;

    /* Setup default values explicitly.
     */
    verbose = 0;
    size_index = 4;		/* Letter paper is default */
    output_name = strdup("/dev/lp0");

    /* Parse command line options.
     */
    while ((c = getopt(argc, argv, "gmw:d:s:vVo:")) != -1)
	switch (c) {
	case 'v':
	    puts("This is oki4drv, version " VERSION ".");
	    exit(EXIT_SUCCESS);
	    break;

	case 'V':
	    verbose = 1;
	    break;

	case 'g':
	    render_mode = graphics;
	    break;

	case 'm':
	    paper_source = manual;
	    break;

	case 'o':
	    {
		free(output_name);
		output_name = strdup(optarg);
	    }
	    break;

	case 's':
	    {		/* paper size */
		struct size_data_t *data = size_data;
		while (data->name && strcmp(data->name, optarg))
		    ++data;
		if (!data->name) {
		    fprintf(stderr, PROGRAM "Incorrect paper size specification: %s",
			    optarg);
		    exit(EXIT_FAILURE);
		} else
		    size_index = data - size_data;
	    }
	    break;

	case 'd':
	    {
		char *result;
		darkness_index = strtol(optarg, &result, 0);
		if (*result != '\0') {
		    fprintf(stderr, PROGRAM "No numeric value given for -d: %s\n",
			    optarg);
		    exit(EXIT_FAILURE);
		}
		if (darkness_index < -2 || darkness_index > 2) {
		    fprintf(stderr, PROGRAM "Darkness (-d) out of range (-2..2).\n");
		    exit(EXIT_FAILURE);
		}
	    }
	    break;

	case 'w':
	    {
		char *result;
		weight_index = strtol(optarg, &result, 0);
		if (*result != '\0') {
		    fprintf(stderr, PROGRAM "No numeric value given for -w: %s\n",
			    optarg);
		    exit(EXIT_FAILURE);
		}
		if (weight_index < -2 || weight_index > 2) {
		    fprintf(stderr, PROGRAM "Darkness (-w) out of range (-2..2).\n");
		    exit(EXIT_FAILURE);
		}
	    }
	    break;

	case '?':
	    if (isprint(optopt))
		fprintf(stderr, PROGRAM "Unknown option `-%c'.\n", optopt);
	    else
		fprintf(stderr, PROGRAM "Unknown option character `\\x%x'.\n", optopt);
	    exit(EXIT_FAILURE);

	default:
	    abort();		/* this shouldn't happen */
	}

    if (verbose) {
	puts("This is oki4drv, version " VERSION ".");
	puts("Copyright (C) 1998,1999,2000 by Marcin Dalecki <dalecki@cs.net.pl>");
	puts("");
	puts("\nRunning in verbose mode...\n");
	printf("output goes to: %s\n", output_name);
	printf("paper size used: %s\n", size_data[size_index].name);
	putchar('\n');
    }

    if (geteuid() != 0) {
	fprintf(stderr, PROGRAM "This program must be SETUID to perform properly!\n");
	exit(EXIT_FAILURE);
    }

    /* Process file by file from the command line.
     */
    for (index = optind; index < argc; index++) {
      if (verbose) {
	printf("processing: %s\n", argv[index]);
      }
      process_file(argv[index]);
    }

    if (verbose)
	puts("Ready!");

    return EXIT_SUCCESS;
}
